"""
Extension to MPL to support the binding of artists to key/mouse events.
"""

import traceback
import sys

from matplotlib import transforms

# CRUFT: matplotlib doesn't yet support canvas.draw_now()
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg


def wxdrawnow(self):
    # print "draw if needed"
    if not self._isDrawn:
        self.draw()


FigureCanvasWxAgg.draw_now = wxdrawnow

if hasattr(transforms, "Transform"):
    # 0.98 transforms
    def pixel_to_data(transform, x, y):
        return transform.inverted().transform_point((x, y))
else:
    # CRUFT 0.91 support
    def pixel_to_data(transform, x, y):
        return transform.inverse_xy_tup((x, y))


def draw_if_needed(canvas):
    """
    Force canvas to be drawn if there are drawing commands outstanding.

    This is needed if the application has emitted a draw_idle() that hasn't
    had time to process prior to the mouse event detection running.
    """
    # Only implemented for wx so far
    if not canvas._isDrawn:
        canvas.draw()


class Selection:
    """
    Store and compare selections.
    """

    # TODO: We need some way to check in prop matches, preferably
    # TODO: without imposing structure on prop.

    artist = None
    prop = {}

    def __init__(self, artist=None, prop={}):
        self.artist, self.prop = artist, self.prop

    def __eq__(self, other):
        return self.artist is other.artist

    def __ne__(self, other):
        return self.artist is not other.artist

    def __bool__(self):
        return self.artist is not None

    __nonzero__ = __bool__  # CRUFT: python 2.x support


#########################################################################
class BindArtist:
    # Track keyboard modifiers for events.
    # TODO: Move keyboard modifier support into the backend.  We cannot
    # TODO: properly support it from outside the windowing system since there
    # TODO: is no way to recognized whether shift is held down when the mouse
    # TODO: first clicks on the the application window.
    _debug = False
    control = False
    shift = False
    alt = False
    meta = False

    # Track doubleclick
    dclick_threshhold = 0.25
    _last_button = None
    _last_time = 0

    # Mouse/keyboard events we can bind to
    events = ["enter", "leave", "motion", "click", "dclick", "drag", "release", "scroll", "key", "keyup"]
    # TODO: Need our own event structure

    def __init__(self, figure):
        self.canvas = figure.canvas
        self.figure = figure

        self._connections = [
            self.canvas.mpl_connect("motion_notify_event", self._onMotion),
            self.canvas.mpl_connect("button_press_event", self._onClick),
            self.canvas.mpl_connect("button_release_event", self._onRelease),
            self.canvas.mpl_connect("key_press_event", self._onKey),
            self.canvas.mpl_connect("key_release_event", self._onKeyRelease),
            self.canvas.mpl_connect("scroll_event", self._onScroll),
        ]

        # Turn off picker if it hasn't already been done
        self.canvas.mpl_disconnect(self.canvas.button_pick_id)
        self.canvas.mpl_disconnect(self.canvas.scroll_pick_id)

        # Clear connections to all artists.
        self.clearall()

    def clear(self, *artists):
        """
        self.clear(h1,h2,...)
            Remove connections for artists h1, h2, ...

        Use clearall() to reset all connections.
        """
        for h in artists:
            for a in self.events:
                if h in self._actions[a]:
                    del self._actions[a][h]
            if h in self._artists:
                self._artists.remove(h)

        if self._current.artist in artists:
            self._current = Selection()
        if self._hasclick.artist in artists:
            self._hasclick = Selection()
        if self._haskey.artist in artists:
            self._haskey = Selection()

    def clearall(self):
        """
        Clear connections to all artists.

        Use clear(h1,h2,...) to reset specific artists.
        """
        # Don't monitor any actions
        self._actions = {}
        for action in self.events:
            self._actions[action] = {}

        # Need activity state
        self._artists = []
        self._current = Selection()
        self._hasclick = Selection()
        self._haskey = Selection()

    def disconnect(self):
        """
        In case we need to disconnect from the canvas...
        """
        try:
            for cid in self._connections:
                self.canvas.mpl_disconnect(cid)
        except:
            pass
        self._connections = []

    def __del__(self):
        self.disconnect()

    def __call__(self, trigger, artist, action):
        """Register a callback for an artist to a particular trigger event.

        usage:
            self.connect(eventname,artist,action)

        where:
            eventname is a string
            artist is the particular graph object to respond to the event
            action(event,**kw) is called when the event is triggered

        The action callback is associated with particular artists.
        Different artists will have different kwargs.  See documentation
        on the contains() method for each artist.  One common properties
        are ind for the index of the item under the cursor, which is
        returned by Line2D and by collections.

        The following events are supported:
            enter: mouse cursor moves into the artist or to a new index
            leave: mouse cursor leaves the artist
            click: mouse button pressed on the artist
            drag: mouse button pressed on the artist and cursor moves
            release: mouse button released for the artist
            key: key pressed when mouse is on the artist
            keyrelease: key released for the artist

        The event received by action has a number of attributes:
            name is the event name which was triggered
            artist is the object which triggered the event
            x,y are the screen coordinates of the mouse
            xdata,ydata are the graph coordinates of the mouse
            button is the mouse button being pressed/released
            key is the key being pressed/released
            shift,control,alt,meta are flags which are true if the
                corresponding key is pressed at the time of the event.
            details is a dictionary of artist specific details, such as the
                id(s) of the point that were clicked.

        When receiving an event, first check the modifier state to be
        sure it applies.  E.g., the callback for 'press' might be:
            if event.button == 1 and event.shift: process Shift-click

        TODO: Only receive events with the correct modifiers (e.g., S-click,
        TODO:   or *-click for any modifiers).
        TODO: Only receive button events for the correct button (e.g., click1
        TODO:   release3, or dclick* for any button)
        TODO: Support virtual artist, so that and artist can be flagged as
        TODO:   having a tag list and receive the correct events
        TODO: Support virtual events for binding to button-3 vs shift button-1
        TODO:   without changing callback code
        TODO: Attach multiple callbacks to the same event?
        TODO: Clean up interaction with toolbar modes
        TODO: push/pushclear/pop context so that binding changes for the duration
        TODO:   e.g., to support ? context sensitive help
        """
        # Check that the trigger is valid
        if trigger not in self._actions:
            raise ValueError("%s invalid --- valid triggers are %s" % (trigger, ", ".join(self.events)))

        # Register the trigger callback
        self._actions[trigger][artist] = action
        # print "==> added",artist,[artist],"to",trigger,":",self._actions[trigger].keys()

        # Maintain a list of all artists
        if artist not in self._artists:
            self._artists.append(artist)

    def trigger(self, actor, action, ev):
        """
        Trigger a particular event for the artist.  Fallback to axes,
        to figure, and to 'all' if the event is not processed.
        """
        if action not in self.events:
            raise ValueError("Trigger expects " + ", ".join(self.events))

        # Tag the event with modifiers
        for mod in ("alt", "control", "shift", "meta"):
            setattr(ev, mod, getattr(self, mod))
        setattr(ev, "artist", None)
        setattr(ev, "action", action)
        setattr(ev, "prop", {})

        # Fallback scheme.  If the event does not return false, pass to parent.
        processed = False
        artist = actor.artist
        prop = actor.prop
        if artist in self._actions[action]:
            # Make sure the x,y data use the coordinate system of the
            # artist rather than the default axes coordinates.
            xy = pixel_to_data(artist.get_transform(), ev.x, ev.y)
            ev.xdata, ev.ydata = xy
            ev.artist = artist
            ev.prop = prop
            processed = self._actions[action][artist](ev)

        if not processed and ev.inaxes in self._actions[action]:
            ev.artist = ev.inaxes
            ev.prop = {}
            processed = self._actions[action][ev.inaxes](ev)

        if not processed and self.figure in self._actions[action]:
            ev.artist = self.figure
            ev.prop = {}
            processed = self._actions[action][self.figure](ev)

        if not processed and "all" in self._actions[action]:
            ev.artist = None
            ev.prop = {}
            processed = self._actions[action]["all"](ev)

        return processed

    def _find_current(self, event):
        """
        Find the artist who will receive the event.  Only search
        registered artists.  All others are invisible to the mouse.
        """
        self.canvas.draw_now()
        # self.canvas.draw()
        # draw_now(self.canvas)
        # TODO: sort by zorder of axes then by zorder within axes
        self._artists.sort(key=lambda x: x.zorder, reverse=True)
        # print "search"," ".join([str(h) for h in self._artists])

        found = Selection()
        # print "searching in",self._artists
        for artist in self._artists:
            # TODO: should contains() return false if invisible?
            if not artist.get_visible():
                continue

            # TODO: optimization - exclude artists not inaxes
            # Note: errors tend to show up on move
            try:
                inside, prop = artist.contains(event)
            except:
                warn_exception("Exception in %s contains() method" % artist)
                inside = False

            if inside:
                found.artist = artist
                found.prop = prop
                break

        # print "found",found.artist

        # TODO: how to check if prop is equal?
        if found != self._current:
            self.trigger(self._current, "leave", event)
            self.trigger(found, "enter", event)
        self._current = found

        return found

    def _onMotion(self, event):
        """
        Track enter/leave/motion through registered artists; all
        other artists are invisible.
        """
        ## Can't kill double-click on motion since Windows produces
        ## spurious motion events.
        # self._last_button = None

        # Dibs on the motion event for the clicked artist
        if self._hasclick:
            self.trigger(self._hasclick, "drag", event)
        else:
            found = self._find_current(event)
            # print "found",found.artist
            self.trigger(found, "motion", event)

    def _onClick(self, event):
        """
        Process button click
        """
        import time

        # Check for double-click
        event_time = time.time()
        # print event_time,self._last_time,self.dclick_threshhold
        # print (event_time > self._last_time + self.dclick_threshhold)
        # print event.button,self._last_button

        if (event.button != self._last_button) or (event_time > self._last_time + self.dclick_threshhold):
            action = "click"
        else:
            action = "dclick"
        self._last_button = event.button
        self._last_time = event_time

        # If an artist is already dragging, feed any additional button
        # presses to that artist.
        # TODO: do we want to force a single button model on the user?
        # TODO: that is, once a button is pressed, no other buttons
        # TODO: can come through?  I think this belongs in canvas, not here.
        if self._hasclick:
            found = self._hasclick
        else:
            found = self._find_current(event)
        # print "button %d pressed"%event.button
        # Note: it seems like if "click" returns False then hasclick should
        # not be set.  The problem is that there are two reasons it can
        # return false: because there is no click action for this artist
        # or because the click action returned false.  A related problem
        # is that click actions will go to the canvas if there is no click
        # action for the artist, even if the artist has a drag. I'll leave
        # it to future maintainers to sort out this problem.  For now the
        # recommendation is that users should define click if they have
        # drag or release on the artist.
        self.trigger(found, action, event)
        self._hasclick = found

    def _onDClick(self, event):
        """
        Process button double click
        """
        # If an artist is already dragging, feed any additional button
        # presses to that artist.
        # TODO: do we want to force a single button model on the user?
        # TODO: that is, once a button is pressed, no other buttons
        # TODO: can come through?  I think this belongs in canvas, not here.
        if self._hasclick:
            found = self._hasclick
        else:
            found = self._find_current(event)
        self.trigger(found, "dclick", event)
        self._hasclick = found

    def _onRelease(self, event):
        """
        Process release release
        """
        self.trigger(self._hasclick, "release", event)
        self._hasclick = Selection()

    def _onKey(self, event):
        """
        Process key click
        """
        # TODO: Do we really want keyboard focus separate from mouse focus?
        # TODO: Do we need an explicit focus command for keyboard?
        # TODO: Can we tab between items?
        # TODO: How do unhandled events get propogated to axes, figure and
        # TODO: finally to application?  Do we need to implement a full tags
        # TODO: architecture a la Tk?
        # TODO: Do modifiers cause a grab?  Does the artist see the modifiers?
        if event.key in ("alt", "meta", "control", "shift"):
            setattr(self, event.key, True)
            return

        if self._haskey:
            found = self._haskey
        else:
            found = self._find_current(event)
        self.trigger(found, "key", event)
        self._haskey = found

    def _onKeyRelease(self, event):
        """
        Process key release
        """
        if event.key in ("alt", "meta", "control", "shift"):
            setattr(self, event.key, False)
            return

        if self._haskey:
            self.trigger(self._haskey, "keyup", event)
        self._haskey = Selection()

    def _onScroll(self, event):
        """
        Process scroll event
        """
        found = self._find_current(event)
        self.trigger(found, "scroll", event)


def warn_exception(msg):
    trace = traceback.format_exception(*sys.exc_info())
    print(msg)
    print(trace[-2][:-1])
    print(trace[-1][:-1])



r"""
Freeform modeling with Chebyshev polynomials

`Chebyshev polynomials <http://en.wikipedia.org/wiki/Chebyshev_polynomials>`_
$T_k$ form a basis set for functions over $[-1, 1]$.  The truncated
interpolating polynomial $P_n$ is a weighted sum of Chebyshev polynomials
up to degree $n$:

.. math::

    f(x) \approx P_n(x) = \sum_{k=0}^n c_i T_k(x)

The interpolating polynomial exactly matches $f(x)$ at the chebyshev
nodes $z_k$ and is near the optimal polynomial approximation to $f$
of degree $n$ under the maximum norm.  For well behaved functions,
the coefficients $c_k$ decrease rapidly, and furthermore are independent
of the degree $n$ of the polynomial.

:class:`FreeformCheby` models the scattering length density profile
of the material within a layer, and :class:`ChebyVF` models the volume
fraction profile of two materials mixed in the layer.

The models can either be defined directly in terms of the Chebyshev
coefficients $c_k$ with *method* = 'direct', or in terms of control
points $(z_k, f(z_k))$ at the Chebyshev nodes :func:`cheby_points`
with *method* = 'interp'.  Bounds on the parameters are easier to
control using 'interp', but the function may oscillate wildly outside
the bounds.  Bounds on the oscillation are easier to control using
'direct', but the shape of the profile is difficult to control.
"""

# TODO: clipping volume fraction to [0, 1] distorts parameter space
# Option 0: clip to [0, 1]
# - Bayesian analysis: parameter values outside the domain will be equally
#   probable out to infinity
# - Newton methods: the fit space is flat outside the domain, which leads
#   to a degenerate hessian.
# - Direct methods: won't fail, but will be subject to random walk
#   performance outside the domain.
# - trivial to implement!
# Option 1: compress (-inf, 0.001] and [0.999, inf) into (0, 0.001], [0.999, 1)
# - won't address any of the problems of clipping
# Option 2: have chisq return inf for points outside the domain
# - Bayesian analysis: correctly assigns probability zero
# - Newton methods: degenerate Hessian outside domain
# - Direct methods: random walk outside domain
# - easy to implement
# Option 3: clip outside domain but add penalty based on amount of clipping
#   A profile based on clipping may have lower chisq than any profile that
#   can be described by a valid model (e.g., by having a sharper transition
#   than would be allowed by the model), leading to a minimum outside D.
#   Adding a penalty constant outside D would help, but there is no constant
#   that works everywhere.  We could use a constant greater than the worst
#   chisq seen so far in D, which can guarantee an arbitrarily low P(x) and
#   a global minimum within D, but for Newton methods, the boundary may still
#   have spurious local minima and objective value now depends on history.
#   Linear compression of profile to fit within the domain would avoid
#   unreachable profile shapes (this is just a linear transform on chebyshev
#   coefficients), and the addition of the penalty value would reduce
#   parameter correlations that result from having transformed parameters
#   resulting in identical profiles.  Returning T = ||A(x)|| from render,
#   with A being a transform that brings the profile within [0, 1], the
#   objective function can return P'(x) = P(x)/(10*(1+sum(T_i)^4) for all
#   slabs i, or P(x) if no slabs return a penalty value.  So long as T is
#   monotonic with increasing badness, with value of 0 within D, and so long
#   as no values of x outside D can generate models that cannot be
#   expressed for any x within D, then any optimizer should return a valid
#   result at the global minimum.  There may still be local minima outside
#   the boundary, so information that the the value is outside the domain
#   still needs to pass through a local optimizer to the fitting program.
#   This approach could be used to transform a box constrained
#   problem to an unconstrained problem using clipping+penalty on the
#   parameter values and removing the need for constrained Newton optimizers.
# - Bayesian analysis: parameters outside D have incorrect probability, but
#   with a sufficiently large penalty, P(x) ~ 0; if the penalty value is
#   too low, details of the correlations outside D may leak into D.
# - Newton methods: Hessian should point back to domain
# - Direct methods: random walk should be biased toward the domain
# - moderately complicated

import numpy as np
from bumps.cheby import cheby_approx, cheby_coeff, cheby_points, cheby_val  # pylint: disable=unused-import
from bumps.parameter import Parameter as Par
from bumps.parameter import to_dict
from numpy import imag, inf, real

from .. import utils
from .layers import Layer


# TODO: add left_sld, right_sld to all layers so that fresnel works
# TODO: access left_sld, right_sld so freeform doesn't need left, right
# TODO: restructure to use vector parameters
# TODO: allow the number of layers to be adjusted by the fit
class FreeformCheby(Layer):
    """
    A freeform section of the sample modeled with Chebyshev polynomials.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    polynomial orders.
    """

    def __init__(self, thickness=0, interface=0, rho=(), irho=(), name="Cheby", method="interp"):
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        self.name = name
        self.method = method
        self.thickness = Par.default(thickness, limits=(0, inf), name=name + " thickness")
        self.rho, self.irho = [
            [Par.default(p, name=name + "[%d] %s" % (i, part), limits=limits) for i, p in enumerate(v)]
            for v, part, limits in zip((rho, irho), ("rho", "irho"), ((-inf, inf), (-inf, inf)))
        ]

    def parameters(self):
        """Return parameters used to define layer"""
        return {"thickness": self.thickness, "rho": self.rho, "irho": self.irho}

    def to_dict(self):
        ret = self.parameters()
        ret.update(
            {
                "type": type(self).__name__,
                "name": self.name,
                "method": self.method,
            }
        )
        return to_dict(ret)

    def render(self, probe, slabs):
        """Render slabs for use with the given probe"""
        thickness = self.thickness.value
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz / thickness
        Prho = _profile([p.value for p in self.rho], t, self.method)
        Pirho = _profile([p.value for p in self.irho], t, self.method)
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)


class ChebyVF(Layer):
    r"""
    Material in a solvent

    :Parameters:

        *thickness* : float | Angstrom
            the thickness of the solvent layer
        *interface* : float | Angstrom
            the rms roughness of the solvent surface
        *material* : Material
            the material of interest
        *solvent* : Material
            the solvent or vacuum
        *vf* : [float]
            the control points for volume fraction
        *method* = 'interp' : string | 'direct' or 'interp'
            freeform profile method

    *method* is 'direct' if the *vf* values refer to chebyshev
    polynomial coefficients or 'interp' if *vf* values refer to
    control points located at $z_k$.

    The control point $k$ is located at $z_k \in [0, L]$ for layer
    thickness $L$, as returned by :func:`cheby_points` called with
    n=len(*vf*) and range=\ $[0, L]$.

    The materials can either use the scattering length density directly,
    such as PDMS = SLD(0.063, 0.00006) or they can use chemical composition
    and material density such as PDMS=Material("C2H6OSi", density=0.965).

    These parameters combine in the following profile formula::

           sld(z) = material.sld * profile(z) + solvent.sld * (1 - profile(z))
    """

    def __init__(self, thickness=0, interface=0, material=None, solvent=None, vf=None, name="ChebyVF", method="interp"):
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        self.name = name
        self.thickness = Par.default(thickness, name="solvent thickness")
        self.interface = Par.default(interface, name="solvent interface")
        self.solvent = solvent
        self.material = material
        self.vf = [Par.default(p, name="vf[%d]" % i) for i, p in enumerate(vf)]
        self.method = method
        # Constraints:
        #   base_vf in [0, 1]
        #   base, length, sigma, thickness, interface>0
        #   base+length+3*sigma <= thickness

    def parameters(self):
        return {
            "solvent": self.solvent.parameters(),
            "material": self.material.parameters(),
            "vf": self.vf,
        }

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "method": self.method,
                "thickness": self.thickness,
                "vf": self.vf,
                "material": self.material,
                "solvent": self.solvent,
            }
        )

    def render(self, probe, slabs):
        Mr, Mi = self.material.sld(probe)
        Sr, Si = self.solvent.sld(probe)
        M = Mr + 1j * Mi
        S = Sr + 1j * Si
        try:
            M, S = M[0], S[0]  # Temporary hack
        except Exception:
            pass

        thickness = self.thickness.value
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz / thickness
        vf = _profile([p.value for p in self.vf], t, self.method)
        vf = np.clip(vf, 0, 1)
        Pw, vf = utils.merge_ends(Pw, vf, tol=1e-3)
        P = M * vf + S * (1 - vf)
        Pr, Pi = real(P), imag(P)
        slabs.extend(rho=[Pr], irho=[Pi], w=Pw)


def _profile(c, t, method):
    r"""
    Evaluate the chebyshev approximation c at points x.

    If method is 'direct' then $c_i$ are the coefficients for the chebyshev
    polynomials $T_i$ yielding $P = \sum_i{c_i T_i(x)}$.

    If method is 'interp' then $c_i$ are the values of the interpolated
    function $f$ evaluated at the chebyshev points returned by
    :func:`cheby_points`.
    """
    if method == "interp":
        c = cheby_coeff(c)
    return cheby_val(c, t)



# standard imports
import json

# third party imports
from bumps.data import parse_multi, strip_quotes
import numpy as np
from orsopy.fileio.orso import load_nexus, load_orso

# refl1d imports
from refl1d.probe import (
    NeutronProbe,
    PolarizedNeutronProbe,
    PolarizedQProbe,
    Probe,
    QProbe,
    XrayProbe,
)
from refl1d.probe.resolution import QL2T, QT2L, FWHM2sigma, dQdL2dT, dQdT2dLoL, sigma2FWHM
from refl1d.sample.reflectivity import BASE_GUIDE_ANGLE


def parse_orso(filename):
    """
    Load an ORSO text (.ort) or binary (.orb) file containing one or more datasets

    Parameters
    ----------
    filename : str
        The path to the ORSO file to be loaded.

    Returns
    -------
    list of tuple
        A list of tuples, each containing a header dictionary and a data array derived from each loaded dataset.
        The header dictionary contains metadata about the measurement,
        and the data array contains the measurement data.

    Notes
    -----
    The function supports both ORSO text (.ort) and binary (.orb) files.
    The polarization information is converted using a predefined mapping.
    The header dictionary includes keys for polarization, angle, angular resolution,
    wavelength, and wavelength resolution.
    """
    if filename.endswith(".ort"):
        entries = load_orso(filename)
    elif filename.endswith(".orb"):
        entries = load_nexus(filename)

    POL_CONVERSION = {
        "po": "++",
        "mo": "--",
        "mm": "--",
        "mp": "-+",
        "pm": "+-",
        "pp": "++",
    }

    entries_out = []
    for entry in entries:
        header = entry.info
        data = entry.data
        settings = header.data_source.measurement.instrument_settings
        columns = header.columns
        polarization = POL_CONVERSION.get(settings.polarization, "unpolarized")
        header_out = {"polarization": polarization}

        def get_key(orso_name, refl1d_name, refl1d_resolution_name):
            """
            Extract value and error from one of the ORSO columns. If no column corresponding
            to entry `orso_name` is found, search in the instrument settings.

            Parameters
            ----------
            orso_name : str
                The name of the ORSO column or instrument setting to extract.
            refl1d_name : str
                The corresponding refl1d name for the value of entry `orso_name`
            refl1d_resolution_name : str
                The corresponding refl1d error name the error of entry `orso_name`

            Notes
            -----
            This function requires the instrument setting `orso_name` to have a "magnitue" and "error" attribute.
            """
            column_index = next(
                (i for i, c in enumerate(columns) if getattr(c, "physical_quantity", None) == orso_name),
                None,
            )
            if column_index is not None:
                # NOTE: this is based on column being second index (under debate in ORSO)
                header_out[refl1d_name] = data[:, column_index]
                cname = columns[column_index].name
                resolution_index = next(
                    (i for i, c in enumerate(columns) if getattr(c, "error_of", None) == cname),
                    None,
                )
                if resolution_index is not None:
                    header_out[refl1d_resolution_name] = data[:, resolution_index]
            else:
                v = getattr(settings, orso_name, None)
                if hasattr(v, "magnitude"):
                    header_out[refl1d_name] = v.magnitude
                if hasattr(v, "error"):
                    header_out[refl1d_resolution_name] = v.error.error_value

        get_key("incident_angle", "angle", "angular_resolution")
        get_key("wavelength", "wavelength", "wavelength_resolution")

        entries_out.append((header_out, np.array(data).T))
    return entries_out


def load4(
    filename,
    keysep=":",
    sep=None,
    comment="#",
    name=None,
    intensity=1,
    background=0,
    back_absorption=1,
    back_reflectivity=False,
    Aguide=BASE_GUIDE_ANGLE,
    H=0,
    theta_offset=None,
    sample_broadening=None,
    L=None,
    dL=None,
    T=None,
    dT=None,
    dR=None,
    FWHM=False,
    radiation=None,
    columns=None,
    data_range=(None, None),
    resolution="normal",
    oversampling=None,
):
    r"""
    Load in four column data Q, R, dR, dQ.

    The file is loaded with *bumps.data.parse_multi*.  *keysep*
    defaults to ':' so that header data looks like JSON key: value
    pairs.  *sep* is None so that the data uses white-space separated
    columns.  *comment* is the standard '#' comment character, used
    for "# key: value" lines, for commenting out data lines using
    "#number number number number", and for adding comments after
    individual data lines.  The parser isn't very sophisticated, so
    be nice.

    *intensity* is the overall beam intensity, *background* is the
    overall background level, and *back_absorption* is the relative
    intensity of data measured at negative Q compared to positive Q
    data.  These can be values or a bumps *Parameter* objects.

    *back_reflectivity* is True if reflectivity was measured through
    the substrate.  This allows you to arrange the model from substrate
    to surface regardless of whether you are measuring through the
    substrate or reflecting off the surface.

    *theta_offset* indicates sample alignment.  In order to use theta
    offset you need to be able to convert from Q to wavelength and angle
    by providing values for the wavelength or the angle, and the associated
    resolution.

    *L*, *dL* in Angstroms can be used to recover angle and angular resolution
    for monochromatic sources where wavelength is fixed and angle is varying.
    These values can also be stored in the file header as::

        # wavelength: 4.75  # Ang
        # wavelength_resolution: 0.02  # Ang (1-sigma)

    *T*, *dT* in degrees can be used to recover wavelength and wavelength
    dispersion for time of flight sources where angle is fixed and wavelength
    is varying, or you can store them in the header of the file::

        # angle: 2  # degrees
        # angular_resolution: 0.2  # degrees (1-sigma)

    If both angle and wavelength are varying in the data, you can specify
    a separate value for each point, such the following::

        # wavelength: [1, 1.2, 1.5, 2.0, ...]
        # wavelength_resolution: [0.02, 0.02, 0.02, ...]

    *dR* can be used to replace the uncertainty estimate for R in the
    file with $\Delta R = R * \text{dR}$.  This allows files with only
    two columns, *Q* and *R* to be loaded.  Note that points with *dR=0*
    are automatically set to the minimum *dR>0* in the dataset.

    Instead of constants, you can provide function, *dT = lambda T: f(T)*,
    *dL = lambda L: f(L)* or *dR = lambda Q, R, dR: f(Q, R, dR)* for more
    complex relationships (with *dR()* returning 1-$\sigma$ $\Delta R$).

    *sample_broadening* in degrees FWHM adds to the angular_resolution.
    Scale 1-$\sigma$ rms by $2 \surd(2 \ln 2) \approx 2.34$ to convert to FWHM.

    *Aguide* and *H* are parameters for polarized beam measurements
    indicating the magnitude and direction of the applied field.

    Polarized data is represented using a multi-section data file,
    with blank lines separating each section.  Each section must
    have a polarization keyword, with value "++", "+-", "-+" or "--".

    *FWHM* is True if dQ, dT, dL are given as FWHM rather than 1-$\sigma$.
    *dR* is always 1-$\sigma$.  *sample_broadening* is always FWHM.

    *radiation* is 'xray' or 'neutron', depending on whether X-ray or
    neutron scattering length density calculator should be used for
    determining the scattering length density of a material.
    Default is 'neutron'

    *columns* is a string giving the column order in the file.  Default
    order is "Q R dR dQ".  Note: include dR and dQ even if the file only
    has two or three columns, but put the missing columns at the end.

    *data_range* indicates which data rows to use.  Arguments are the
    same as the list slice arguments, *(start, stop, step)*.  This follows
    the usual semantics of list slicing, *L[start:stop:step]*, with
    0-origin indices, stop is last plus one and step optional.  Use negative
    numbers to count from the end.  Default is *(None, None)* for the entire
    data set.

    *resolution* is 'normal' (default) or 'uniform'. Use uniform if you
    are merging Q points from a finely stepped energy sensitive measurement.

    *oversampling* is None or a positive integer indicating how many points to add
    between data point to support sparse data with denser theory (for PolarizedNeutronProbe)
    """
    json_header_encoding = False

    if filename.endswith(".ort") or filename.endswith(".orb"):
        entries = parse_orso(filename)
    else:
        json_header_encoding = True  # for .refl files, header values are json-encoded
        entries = parse_multi(filename, keysep=keysep, sep=sep, comment=comment)

    if columns:
        actual = columns.split()
        natural = "Q R dR dQ".split()
        column_order = [actual.index(k) for k in natural]
    else:
        column_order = [0, 1, 2, 3]
    index = slice(*data_range)
    probe_args = dict(
        name=name,
        filename=filename,
        intensity=intensity,
        background=background,
        back_absorption=back_absorption,
        back_reflectivity=back_reflectivity,
        theta_offset=theta_offset,
        sample_broadening=sample_broadening,
        resolution=resolution,
    )
    data_args = dict(
        radiation=radiation,
        FWHM=FWHM,
        T=T,
        L=L,
        dT=dT,
        dL=dL,
        dR=dR,
        column_order=column_order,
        index=index,
    )
    if len(entries) == 1:
        probe = _data_as_probe(entries[0], json_header_encoding, probe_args, **data_args)
    else:
        data_by_xs = {
            strip_quotes(entry[0]["polarization"]): _data_as_probe(entry, json_header_encoding, probe_args, **data_args)
            for entry in entries
        }
        if not set(data_by_xs.keys()) <= set("-- -+ +- ++".split()):
            raise ValueError("Unknown cross sections in: " + ", ".join(sorted(data_by_xs.keys())))
        xs = [data_by_xs.get(xs, None) for xs in ("--", "-+", "+-", "++")]

        if any(isinstance(d, QProbe) for d in xs if d is not None):
            probe = PolarizedQProbe(xs, Aguide=Aguide, H=H)
        else:
            probe = PolarizedNeutronProbe(xs, Aguide=Aguide, H=H, oversampling=oversampling)
    return probe


def _data_as_probe(
    entry,
    json_header_encoding,
    probe_args,
    T,
    L,
    dT,
    dL,
    dR,
    FWHM,
    radiation,
    column_order,
    index,
):
    decoder = json.loads if json_header_encoding else lambda x: x
    name = probe_args["filename"]
    header, data = entry
    if len(data) == 2:
        data_Q, data_R = (data[k][index] for k in column_order[:2])
        data_dR, data_dQ = None, None
        if dR is None:
            raise ValueError("Need dR for two column data in %r" % name)
    elif len(data) == 3:
        data_Q, data_R, data_dR = (data[k][index] for k in column_order[:3])
        data_dQ = None
    else:
        data_Q, data_R, data_dR, data_dQ = (data[k][index] for k in column_order)

    if FWHM and data_dQ is not None:  # dQ is already 1-sigma when not FWHM
        data_dQ = FWHM2sigma(data_dQ)

    # Override dR in the file if desired.
    # Make sure the computed dR is positive (otherwise chisq is infinite) by
    # choosing the smallest positive uncertainty to replace the invalid values.
    if dR is not None:
        data_dR = dR(data_Q, data_R, data_dR) if callable(dR) else data_R * dR
        data_dR[data_dR <= 0] = np.min(data_dR[data_dR > 0])

    # support calculation of sld from material based on radiation type
    if radiation is not None:
        data_radiation = radiation
    elif "radiation" in header:
        data_radiation = decoder(header["radiation"])
    else:
        # Default to neutron data if radiation not given in head.
        data_radiation = "neutron"
        # data_radiation = None

    if data_radiation == "xray":
        make_probe = XrayProbe
    elif data_radiation == "neutron":
        make_probe = NeutronProbe
    else:
        make_probe = Probe

    # Get T,dT,L,dL from header if it is not provided as an argument
    def fetch_key(key, override):
        # Note: pulls header and index pulled from context
        if override is not None:
            return override
        elif key in header:
            v = decoder(header[key])
            return np.array(v)[index] if isinstance(v, list) else v
        else:
            return None

    # Get T and L, either from user input or from datafile.
    data_T = fetch_key("angle", T)
    data_L = fetch_key("wavelength", L)

    # If one of T and L is missing, reconstruct it from Q
    if data_T is None and data_L is not None:
        data_T = QL2T(data_Q, data_L)
    if data_L is None and data_T is not None:
        data_L = QT2L(data_Q, data_T)

    # Get dT and dL, either from user input or from datafile.
    data_dL = fetch_key("wavelength_resolution", dL)
    data_dT = fetch_key("angular_resolution", dT)
    # print(header['angular_resolution'], data_dT)

    # Support dT = f(T), dL = f(L)
    if callable(data_dT):
        if data_T is None:
            raise ValueError("Need T to determine dT for %r" % name)
        data_dT = data_dT(data_T)
    if callable(data_dL):
        if data_L is None:
            raise ValueError("Need L to determine dL for %r" % name)
        data_dL = data_dL(data_L)

    # Convert input dT,dL to FWHM if necessary.
    if data_dL is not None and not FWHM:
        data_dL = sigma2FWHM(data_dL)
    if data_dT is not None and not FWHM:
        data_dT = sigma2FWHM(data_dT)

    # If dT or dL is missing, reconstruct it from Q.
    if data_dT is None and not any(v is None for v in (data_L, data_dL, data_dQ)):
        data_dT = dQdL2dT(data_Q, data_dQ, data_L, data_dL)
    if data_dL is None and not any(v is None for v in (data_T, data_dT, data_dQ)):
        data_dLoL = dQdT2dLoL(data_Q, data_dQ, data_T, data_dT)
        data_dL = data_dLoL * data_L

    # Check reconstruction if user provided any of T, L, dT, or dL.
    # Also, sample_offset or sample_broadening.
    offset = probe_args["theta_offset"]
    broadening = probe_args["sample_broadening"]
    if any(v is not None for v in (T, dT, L, dL, offset, broadening)):
        if data_T is None:
            raise ValueError("Need L to determine T from Q for %r" % name)
        if data_L is None:
            raise ValueError("Need T to determine L from Q for %r" % name)
        if data_dT is None:
            raise ValueError("Need dL to determine dT from dQ for %r" % name)
        if data_dL is None:
            raise ValueError("Need dT to determine dL from dQ for %r" % name)

    # Build the probe, or the Q probe if we don't have angle and wavelength.
    if all(v is not None for v in (data_T, data_L, data_dT, data_dL)):
        probe = make_probe(
            T=data_T,
            dT=data_dT,
            L=data_L,
            dL=data_dL,
            data=(data_R, data_dR),
            dQo=data_dQ,
            **probe_args,
        )
    else:
        # QProbe doesn't accept theta_offset or sample_broadening
        qprobe_args = probe_args.copy()
        qprobe_args.pop("theta_offset")
        qprobe_args.pop("sample_broadening")
        probe = QProbe(data_Q, data_dQ, data=(data_R, data_dR), **qprobe_args)

    return probe



import matplotlib.pyplot as plt
import wx
from bumps.fitproblem import FitProblem
from bumps.gui.util import EmbeddedPylab
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as Toolbar

# The Figure object is used to create backend-independent plot representations.
from matplotlib.figure import Figure
from numpy import inf

from refl1d.probe import Probe

# Can't seem to detect when notebook should be drawn on Mac
IS_MAC = wx.Platform == "__WXMAC__"


# ------------------------------------------------------------------------
class DataView(wx.Panel):
    title = "Reflectivity"
    default_size = (600, 400)

    def __init__(self, *args, **kw):
        wx.Panel.__init__(self, *args, **kw)

        # Instantiate a figure object that will contain our plots.
        figure = Figure(figsize=(1, 1), dpi=72)

        # Initialize the figure canvas, mapping the figure object to the plot
        # engine backend.
        canvas = FigureCanvas(self, wx.ID_ANY, figure)

        # Wx-Pylab magic ...
        # Make our canvas an active figure manager for pylab so that when
        # pylab plotting statements are executed they will operate on our
        # canvas and not create a new frame and canvas for display purposes.
        # This technique allows this application to execute code that uses
        # pylab stataments to generate plots and embed these plots in our
        # application window(s).  Use _activate_figure() to set.
        self.pylab_interface = EmbeddedPylab(canvas)

        # Instantiate the matplotlib navigation toolbar and explicitly show it.
        mpl_toolbar = Toolbar(canvas)
        mpl_toolbar.Realize()

        # Create a vertical box sizer to manage the widgets in the main panel.
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(canvas, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, border=0)
        sizer.Add(mpl_toolbar, 0, wx.EXPAND | wx.ALL, border=0)

        # Associate the sizer with its container.
        self.SetSizer(sizer)
        sizer.Fit(self)

        self.view = Probe.view

        self._need_redraw = False
        self.Bind(wx.EVT_SHOW, self.OnShow)
        self._calculating = False
        self.toolbar = mpl_toolbar

    def menu(self):
        # Add 'View' menu to the menu bar and define its options.
        # Present y-axis plotting scales as radio buttons.
        # Grey out items that are not currently implemented.
        frame = wx.GetTopLevelParent(self)
        menu = wx.Menu()
        _item = menu.AppendRadioItem(wx.ID_ANY, "&Fresnel", "Plot R/R_F")
        frame.Bind(wx.EVT_MENU, self.OnFresnel, _item)
        _item.Check(Probe.view == "fresnel")
        _item = menu.AppendRadioItem(wx.ID_ANY, "Log Fresnel", "Plot log R/R_F")
        frame.Bind(wx.EVT_MENU, self.OnLogFresnel, _item)
        _item.Check(Probe.view == "logfresnel")
        _item = menu.AppendRadioItem(wx.ID_ANY, "Li&near", "Plot linear R")
        frame.Bind(wx.EVT_MENU, self.OnLinear, _item)
        _item.Check(Probe.view == "linear")
        _item = menu.AppendRadioItem(wx.ID_ANY, "&Log", "Plot log R")
        frame.Bind(wx.EVT_MENU, self.OnLog, _item)
        _item.Check(Probe.view == "log")
        _item = menu.AppendRadioItem(wx.ID_ANY, "&Q4", "Plot R * Q^4")
        frame.Bind(wx.EVT_MENU, self.OnQ4, _item)
        _item.Check(Probe.view == "q4")
        _item = menu.AppendRadioItem(wx.ID_ANY, "&SA", "Plot spin asymmetry")
        frame.Bind(wx.EVT_MENU, self.OnSA, _item)
        _item.Check(Probe.view == "SA")

        menu.AppendSeparator()

        _item = menu.Append(wx.ID_ANY, "&Residuals", "Plot residuals (R_theory - R)/dR")
        frame.Bind(wx.EVT_MENU, self.OnResiduals, _item)
        menu.Enable(id=_item.GetId(), enable=True)

        return menu

    # ==== Views ====
    def OnLog(self, event):
        self.view = "log"
        self.redraw(reset=True)

    def OnLinear(self, event):
        self.view = "linear"
        self.redraw(reset=True)

    def OnFresnel(self, event):
        self.view = "fresnel"
        self.redraw(reset=True)

    def OnLogFresnel(self, event):
        self.view = "logfresnel"
        self.redraw(reset=True)

    def OnQ4(self, event):
        self.view = "q4"
        self.redraw(reset=True)

    def OnSA(self, event):
        self.view = "SA"
        self.redraw(reset=True)

    def OnResiduals(self, event):
        self.view = "residuals"
        self.redraw(reset=True)

    # ==== Model view interface ===
    def OnShow(self, event):
        # print "theory show"
        if not event.Show:
            return
        # print "showing theory"
        if self._need_redraw:
            # print "-redraw"
            self.redraw()

    def get_state(self):
        return self.problem

    def set_state(self, state):
        self.set_model(state)

    def set_model(self, model):
        # print ">>>>>>> refl1d data set model"
        self.problem = model
        self.redraw(reset=True)

    def update_model(self, model):
        # print ">>>>>>> refl1d data update model"
        if self.problem == model:
            self.redraw()

    def update_parameters(self, model):
        # print ">>>>>>> refl1d data update parameters"
        if self.problem == model:
            self.redraw()

    # =============================

    def redraw(self, reset=False):
        # Hold off drawing until the tab is visible
        if not IS_MAC and not self.IsShown():
            self._need_redraw = True
            return
        # print "drawing theory"

        if self._calculating:
            # That means that I've entered the thread through a
            # wx.Yield for the currently executing redraw.  I need
            # to cancel the running thread and force it to start
            # the calculation over.
            self._cancel_calculate = True
            # print "canceling calculation"
            return

        self._need_redraw = False
        self._calculating = True

        # Calculate reflectivity
        # print "calling again"
        while True:
            # print "restarting"
            # We are restarting the calculation, so clear the reset flag
            self._cancel_calculate = False

            # Preform the calculation
            if isinstance(self.problem, FitProblem):
                # print "n=",len(self.problem.models)
                for p in self.problem.models:
                    if hasattr(p, "reflectivity"):
                        self._precalc(p)
                        # print "cancel",self._cancel_calculate,"reset",p.fitness.is_reset()
                        if p.is_reset() or self._cancel_calculate:
                            break
                if self._cancel_calculate or self.problem.active_model.is_reset():
                    continue
            else:
                self._precalc(self.problem)
                if self._cancel_calculate or self.problem.is_reset():
                    continue

            # Redraw the canvas with newly calculated reflectivity
            with self.pylab_interface:
                ax = plt.gca()
                # print "reset",reset, ax.get_autoscalex_on(), ax.get_xlim()
                range_x = ax.get_xlim()
                range_y = ax.get_ylim()
                # print "composing"
                plt.clf()  # clear the canvas
                # shift=20 if self.view == 'log' else 0
                shift = 0
                if isinstance(self.problem, FitProblem):
                    for _, p in enumerate(self.problem.models):
                        if hasattr(p, "reflectivity"):
                            p.plot_reflectivity(view=self.view, plot_shift=shift)
                            if self._cancel_calculate or p.is_reset():
                                break
                    if self._cancel_calculate or self.problem.active_model.is_reset():
                        continue
                else:
                    self.problem.plot_reflectivity(view=self.view, plot_shift=shift)
                    if self._cancel_calculate or self.problem.is_reset():
                        continue

                try:
                    # If we can calculate chisq, then put it on the graph.
                    text = "chisq=" + self.problem.chisq_str()
                    plt.text(0.01, 0.01, text, transform=plt.gca().transAxes)
                except Exception:
                    pass
                # print "drawing"
                if not reset:
                    self.toolbar.push_current()
                    ax = plt.gca()
                    ax.set_xlim(range_x)
                    ax.set_ylim(range_y)
                    self.toolbar.push_current()
                plt.draw()
                # print "done drawing"
                break

        self._calculating = False

    def _precalc(self, fitness):
        # First calculate reflectivity
        fitness.reflectivity(resolution=False)
        # print "yield 1"
        wx.Yield()
        if self._cancel_calculate or fitness.is_reset():
            return
        # Then calculate resolution
        fitness.reflectivity()
        # print "yield 2"
        wx.Yield()
        if self._cancel_calculate or fitness.is_reset():
            return


# DEPRECATED - using fixed y-range
def set_xrange(ax, range_x):
    miny, maxy = inf, -inf
    for L in ax.get_lines():
        x, y = L.get_data()
        idx = (x > range_x[0]) & (x < range_x[1])
        if idx.any():
            miny = min(miny, min(y[idx]))
            maxy = max(maxy, max(y[idx]))
    if miny < maxy:
        if ax.get_yscale() == "linear":
            padding = 0.05 * (maxy - miny)
            miny, maxy = miny - padding, maxy + padding
        else:
            miny, maxy = miny * 0.95, maxy * 1.05
    ax.set_xlim(range_x)
    ax.set_ylim(miny, maxy)



from numpy import fabs, sqrt, exp


def refl(layers, kz, depth, sigma, rho, irho):
    J = 1j

    # // Check that Q is not too close to zero.
    # // For negative Q, reverse the layers.
    cutoff = 1e-10
    sigma_offset = 0
    if kz >= cutoff:
        i_next = 0
        step = 1
    elif kz <= -cutoff:
        i_next = layers - 1
        step = -1
        sigma_offset = -1
    else:
        return complex(-1, 0)

    # // Since sqrt(1/4 * x) = sqrt(x)/2, I'm going to pull the 1/2 into the
    # // sqrt to save a multiplication later.
    pi4 = 12.566370614359172e-6  # // 1e-6 * 4 pi
    kz_sq = kz * kz + pi4 * rho[i_next]  # // kz^2 + 4 pi Vrho
    k = fabs(kz)

    B11 = B22 = 1
    B12 = B21 = 0

    for i in range(layers - 1):
        # // The loop index is not the layer number because we may be reversing
        # // the stack.  Instead, n is set to the incident layer (which may be
        # // first or last) and incremented or decremented each time through.
        k_next = sqrt(kz_sq - pi4 * complex(rho[i_next + step], irho[i_next + step]))
        F = (k - k_next) / (k + k_next) * exp(-2.0 * k * k_next * sigma[sigma_offset + i_next] ** 2)
        M11 = exp(J * k * depth[i_next]) if i > 0 else 1.0
        M22 = exp(-J * k * depth[i_next]) if i > 0 else 1.0
        M21 = F * M11
        M12 = F * M22

        # // Multiply existing layers B by new layer M
        # // We have unrolled the matrix multiply for speed.
        C1 = B11 * M11 + B21 * M12
        C2 = B11 * M21 + B21 * M22
        B11 = C1
        B21 = C2
        C1 = B12 * M11 + B22 * M12
        C2 = B12 * M21 + B22 * M22
        B12 = C1
        B22 = C2
        i_next += step
        k = k_next

    # // And we are done.
    return B12 / B11


def reflectivity_amplitude(depth, sigma, rho, irho, kz, rho_index, r):
    layers = len(depth)
    points = len(kz)
    for i in range(points):
        offset = rho_index[i]
        r[i] = refl(layers, kz[i], depth, sigma, rho[offset], irho[offset])



# Author: Paul Kienzle
"""
Convert staj files to refl1d
"""

import os

import numpy as np
from bumps import parameter
from numpy import cos, degrees, pi, radians, sqrt, tan

from ...probe.resolution import QL2T, sigma2FWHM
from ...sample.layers import Repeat, Slab, Stack
from ...sample.magnetism import Magnetism
from ...sample.material import SLD
from ..probe import NeutronProbe, PolarizedNeutronProbe, XrayProbe
from .staj import MlayerMagnetic, MlayerModel


def load_mlayer(filename, fit_pmp=0, name=None, layers=None):
    """
    Load a staj file as a model.
    """
    if filename.endswith(".staj"):
        staj = MlayerModel.load(filename)
        model = mlayer_to_model(staj, name=name, layers=layers)
    else:
        sta = MlayerMagnetic.load(filename)
        model = mlayer_magnetic_to_model(sta, name=name, layers=layers)
    if fit_pmp != 0:
        fit_all(model, pmp=fit_pmp)
    return model


def save_mlayer(experiment, filename, datafile=None):
    """
    Save a model to a staj file.
    """
    staj = model_to_mlayer(experiment, datafile)
    # print staj
    staj.save(filename)


def fit_all(M, pmp=20):
    """
    Set all non-zero parameters to fitted parameters inside the model.
    """

    # Exclude unlikely fitting parameters
    exclude = set(
        (
            M.sample[0].thickness,
            M.sample[-1].thickness,
            M.sample[-1].interface,
            M.probe.back_absorption,
        )
    )
    if M.probe.intensity.value == 1:
        exclude.add(M.probe.intensity)
    if M.probe.background.value < 2e-10:
        exclude.add(M.probe.background.value)

    # Fit everything else using a range of +/- pmp %
    for p in parameter.unique(M.parameters()):
        if p in exclude:
            continue
        if p.value != 0:
            p.pmp(pmp)
        # p.fixed = False


def mlayer_to_model(staj, name=None, layers=None):
    """
    Convert a loaded staj file to a refl1d experiment.

    Returns a new experiment
    """
    from ...experiment import Experiment

    sample = _mlayer_to_stack(staj, name, layers)
    probe = _load_probe(staj, name, xs="")
    return Experiment(sample=sample, probe=probe)


def _mlayer_to_stack(s, name, layers):
    """
    Return a sample stack based on the model used in the staj file.
    """
    # check pars
    if layers and len(layers) != len(s.rho):
        raise ValueError("Have %d staj layers but got %d layer names" % (len(s.rho), len(layers)))

    # prepend datafile name to layers
    if name is None:
        name = os.path.splitext(s.data_file)[0]
    if name and not name.endswith(" "):
        name += " "

    i1 = s.num_top + 1
    i2 = s.num_top + s.num_middle + 1

    # Construct slabs
    slabs = []
    for i in reversed(range(len(s.rho))):
        if layers:
            Lname = layers[len(layers) - i - 1]
        elif i == 0:
            Lname = "V"
        elif i < i1:
            Lname = "T%d" % (i)
        elif i < i2:
            Lname = "M%d" % (i - i1 + 1)
        else:
            Lname = "B%d" % (i - i2 + 1)
        slabs.append(
            Slab(
                material=SLD(rho=s.rho[i], irho=s.irho[i], name=name + Lname),
                thickness=s.thickness[i],
                interface=s.sigma_roughness[i],
            )
        )

    # Build stack
    if s.num_repeats == 0:
        stack = Stack(slabs[:i1] + slabs[i2:])
    elif s.num_repeats == 1:
        stack = Stack(slabs)
    else:
        stack = Stack(slabs[:i1] + [Repeat(Stack(slabs[i1:i2]))] + slabs[i2:])

    return stack


XS = {"A": "--", "B": "-+", "C": "+-", "D": "++", "": ""}


def _load_probe(s, name, xs):
    if name is None:
        name = os.path.splitext(s.data_file)[0]

    if s.data_file == "":
        filename = "simulated"
        Q = np.linspace(s.Qmin, s.Qmax, s.num_Q)
        R, dR = None, None
    else:
        filename = s.data_file
        Q, R, dR = np.loadtxt(s.data_file + xs).T

    # Use Q and wavelength L from the staj file to determine angle T
    L = s.wavelength
    T = QL2T(Q=Q, L=s.wavelength)

    # Refl1D uses the following for dQ:
    #
    #     (dQ/Q)^2 = (dL/L)^2 + (dT/tan(T))^2
    #
    # Given the mlayer resolution dQ, and wavelength divergence dL, we can solve the
    # above for dT such that dQ in refl1d matches dQ in mlayer.
    #
    #      dT = sqrt( tan(T)^2 * ( (dQ/Q)^2 - (dL/L)^2) ) )
    #         = sqrt( (dQ*L/4*pi*sin(T))^2*(sin(T)/cos(T))^2 - (tan(T)*dL/L)^2 )
    #         = sqrt( (dQ*L/4*pi*cos(T))^2 - (tan(T)*dL/L)^2 )
    dQ = s.FWHMresolution(Q)
    dL = s.wavelength_dispersion
    dT = degrees(sqrt((dQ * L / (4 * pi * cos(radians(T)))) ** 2 - (tan(radians(T)) * dL / L) ** 2))

    # Hack: X-ray is 1.54; anything above 2 is neutron.  Doesn't matter since
    # materials are set as SLD rather than composition.
    if s.wavelength < 2:
        probe = XrayProbe
    else:
        probe = NeutronProbe
    probe = probe(
        T=T,
        dT=dT,
        L=L,
        dL=dL,
        data=(R, dR),
        theta_offset=getattr(s, "theta_offset", 0),  # gj2 has no theta offset
        background=s.background,
        intensity=s.intensity,
        name=name,
    )
    probe.filename = filename
    # probe.oversample(n=10)
    return probe


def model_to_mlayer(model, datafile):
    """
    Return an mlayer model based on the a slab stack.

    Raises TypeError if model cannot be stored as a staj file.
    """
    stack = model.sample
    probe = model.probe

    staj = MlayerModel(roughness_steps=51)

    # Set up beam info
    if (probe.L != probe.L[0]).any():
        # Reason is that mlayer uses mu/(2 lambda) rather than irho
        raise TypeError("Mlayer only supports monochromatic sources")

    staj.set(
        wavelength=probe.L[0],
        intensity=probe.intensity.value,
        background=probe.background.value,
        theta_offset=probe.theta_offset.value,
    )
    if datafile:
        staj.data_file = os.path.basename(datafile)
    else:
        staj.Qmin, staj.Qmax = min(probe.Qo), max(probe.Qo)
        staj.num_Q = len(probe.Qo)
    staj.fit_FWHMresolution(probe.Qo, sigma2FWHM(probe.dQ))

    # Interpret slabs and repeats
    sections = []
    section = []
    repeats = 0
    for l in stack:
        if isinstance(l, Slab):
            section.append(l)
        elif isinstance(l, Repeat):
            sections.append(section)
            sections.append(l[:])
            repeats = l.repeat
            section = []
        else:
            raise TypeError("Only slabs supported")
    sections.append(section)
    if len(sections) > 3:
        raise TypeError("Only one repeated section supported")
    if len(sections) == 3:
        for l in sections[1]:
            if not isinstance(l, Slab):
                raise TypeError("Only slabs supported in repeat section")
        num_top = len(sections[2]) - 1
        num_middle = len(sections[1])
        num_bottom = len(sections[0])
        if num_top > 9 or num_middle > 9 or num_bottom > 9:
            raise TypeError("Maximum section length of 9")
        if num_top < 1 or num_middle < 1 or num_bottom < 1:
            raise TypeError("Need at least one slab per section, plus vacuum")
        staj.num_top = num_top
        staj.num_middle = num_middle
        staj.num_bottom = num_bottom
        staj.num_repeats = repeats
        slabs = []
        slabs.extend(reversed(sections[2]))
        slabs.extend(reversed(sections[1]))
        slabs.extend(reversed(sections[0]))
    else:
        # must be only one section
        slabs = list(reversed(sections[0]))
        if len(slabs) > 28:
            raise TypeError("Too many slabs (only 28 slabs allowed)")

    # Convert slabs to sld parameters
    # print "\n".join(str(s) for s in slabs)
    values = []
    for layer in slabs:
        rho, irho = layer.material.sld(probe)
        thickness = layer.thickness.value
        roughness = layer.interface.value
        values.append((rho, irho, thickness, roughness))
    vectors = [np.array(v) for v in zip(*values)]

    # If back reflectivity, reverse the layers and move the interfaces
    # from the top of the layer to the bottom.
    if probe.back_reflectivity:
        vectors = [v[::-1] for v in vectors]
        vectors[3] = np.roll(vectors[3], 1)
        staj.num_top, staj.num_bottom = staj.num_bottom, staj.num_top

    staj.rho, staj.irho, staj.thickness, staj.sigma_roughness = vectors

    # If no repeats, split the model into sections
    if len(sections) == 1:
        # If the stack is too short, add layers at the top until it is
        # tall enough.  These layers should have the same SLD as the
        # old top layer, and a thickness large enough to accommodate
        # the interface between the top layer and the second layer.
        while len(staj.rho) < 4:
            staj.rho = np.hstack((staj.rho[0], staj.rho))
            staj.irho = np.hstack((staj.irho[0], staj.irho))
            staj.thickness = np.hstack((0, 3.5 * staj.sigma_roughness[1], staj.thickness[1:]))
            staj.sigma_roughness = np.hstack((0, staj.sigma_roughness))

        staj.split_sections()

    return staj


def mlayer_magnetic_to_model(sta, name=None, layers=None):
    """
    Convert a loaded sta file to a refl1d experiment.

    Returns a new experiment
    """
    from ...experiment import Experiment

    sample = _mlayer_magnetic_to_stack(sta, name, layers)
    probe = _mlayer_magnetic_to_probe(sta, name)
    return Experiment(sample=sample, probe=probe, dz=0.1)


def _mlayer_magnetic_to_stack(s, name, layers):
    """
    Return a sample stack based on the model used in the sta file.
    """
    # check pars
    if layers and len(layers) != len(s.rho):
        raise ValueError("Have %d sta layers but got %d layer names" % (len(s.rho), len(layers)))

    # prepend datafile name to layers
    if name is None:
        name = os.path.splitext(s.data_file)[0]
    if name and not name.endswith(" "):
        name += " "

    # Construct slabs
    magnetic_offset = np.cumsum(s.thickness - s.mthickness)
    slabs = []
    nlayers = len(s.rho)
    for i in range(nlayers - 1, -1, -1):
        if layers:
            Lname = layers[len(layers) - i - 1]
        elif i == 0:
            Lname = "V"
        else:
            Lname = "M%d" % i
        slab_i = Slab(
            material=SLD(rho=s.rho[i], irho=s.irho[i], name=name + Lname),
            thickness=s.thickness[i],
            interface=s.sigma_roughness[i],
        )
        if s.mrho[i] != 0.0:
            slab_i.magnetism = Magnetism(
                s.mrho[i],
                s.mtheta[i],
                interface_below=s.sigma_mroughness[i + 1] if i < nlayers - 1 else 0,
                interface_above=s.sigma_mroughness[i],
                dead_below=magnetic_offset[i + 1] if i < nlayers - 1 else 0,
                dead_above=magnetic_offset[i],
            )
        slabs.append(slab_i)

    return Stack(slabs)


def _mlayer_magnetic_to_probe(s, name):
    """
    Return a model probe based on the data used for the staj file.
    """
    if name is None:
        name = os.path.splitext(s.data_file)[0]

    active_xsec = s.active_xsec.upper()
    xs = [_load_probe(s, name, xs) if (xs in active_xsec) else None for xs in "ABCD"]
    probe = PolarizedNeutronProbe(xs, Aguide=s.guide_angle)
    # probe.oversample(n=6)
    return probe



import inspect

from bumps.parameter import Calculation, Parameter, to_dict
from numpy import asarray, broadcast_to, imag, real

from refl1d import utils
from .layers import Layer
from .magnetism import DEFAULT_THETA_M, BaseMagnetism, Magnetism
from .material import SLD


class FunctionalProfile(Layer):
    """
    Generic profile function

    Parameters:

        *thickness* the thickness of the layer

        *interface* the roughness of the surface [not implemented]

        *profile* the profile function, suitably parameterized

        *tol* is the tolerance for considering values equal

        *magnetism* magnetic profile associated with the layer

        *name* is the layer name

    The profile function takes a depth vector *z* returns a density vector
    *rho*. For absorbing profiles, return complex vector *rho + irho*1j*.

    Fitting parameters are the available named arguments to the function.
    The first argument is a depth vector, which is the array of depths at
    which the profile is to be evaluated.  It is guaranteed to be increasing,
    with step size 2*z[0].

    Initial values for the function parameters can be given using name=value.
    These values can be scalars or fitting parameters.  The function will
    be called with the current parameter values as arguments.  The layer
    thickness can be computed as :func:`layer_thickness`.

    There is no mechanism for querying the larger profile to determine the
    value of the *rho* at the layer boundaries.  If needed, this information
    will have to be communicated through shared parameters.  For example::

        L1 = SLD('L1', rho=2.07)
        L3 = SLD('L3', rho=4)
        def linear(z, rhoL, rhoR):
            rho = z * (rhoR-rhoL)/(z[-1]-z[0]) + rhoL
            return rho
        profile = FunctionalProfile(100, 0, profile=linear,
                                    rhoL=L1.rho, rhoR=L3.rho)
        sample = L1 | profile | L3
    """

    RESERVED = ("thickness", "interface", "profile", "tol", "magnetism", "name")

    # TODO: test that thickness(z) matches the thickness of the layer
    def __init__(self, thickness=0, interface=0, profile=None, tol=1e-3, magnetism=None, name=None, **kw):
        if not name:
            name = profile.__name__
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        if profile is None:
            raise TypeError("Need profile")
        self.name = name
        self.thickness = Parameter.default(thickness, name=name + " thickness")
        self.interface = Parameter.default(interface, name=name + " interface")
        self.profile = profile
        self.tol = tol
        self.magnetism = magnetism

        # TODO: maybe make these lazy (and for magnetism below as well)
        rho_start = _LayerLimit(self, isend=False, isrho=True)
        irho_start = _LayerLimit(self, isend=False, isrho=False)
        rho_end = _LayerLimit(self, isend=True, isrho=True)
        irho_end = _LayerLimit(self, isend=True, isrho=False)
        self.start = SLD(name + " start", rho=rho_start, irho=irho_start)
        self.end = SLD(name + " end", rho=rho_end, irho=irho_end)

        _set_parameters(self, name, profile, kw, self.RESERVED)

    def parameters(self):
        P = _get_parameters(self)
        P["thickness"] = self.thickness
        # P['interface'] = self.interface
        return P

    def to_dict(self):
        return to_dict(
            {
                "type": type(self).__name__,
                "name": self.name,
                "thickness": self.thickness,
                "interface": self.interface,
                "profile": self.profile,
                "parameters": _get_parameters(self),
                "tol": self.tol,
                "magnetism": self.magnetism,
            }
        )

    def render(self, probe, slabs):
        Pw, Pz = slabs.microslabs(self.thickness.value)
        if len(Pw) == 0:
            return
        # print kw
        # TODO: always return rho, irho from profile function
        # return value may be a constant for rho or irho
        phi = asarray(self.profile(Pz, **_get_values(self)))
        if phi.shape != Pz.shape:
            raise TypeError("profile function '%s' did not return array phi(z)" % self.profile.__name__)
        Pw, phi = utils.merge_ends(Pw, phi, tol=self.tol)
        # P = M*phi + S*(1-phi)
        slabs.extend(rho=[real(phi)], irho=[imag(phi)], w=Pw)
        # slabs.interface(self.interface.value)


class FunctionalMagnetism(BaseMagnetism):
    """
    Functional magnetism profile.

    Parameters:

        *profile* the profile function, suitably parameterized

        *tol* is the tolerance for considering values equal

        :class:`refl1d.sample.magnetism.BaseMagnetism` parameters

    The profile function takes a depth vector *z* and returns a magnetism
    vector *rhoM*. For magnetic twist, return a pair of vectors *(rhoM, thetaM)*.
    Constants can be returned for *rhoM* or *thetaM*.  If *thetaM* is not
    provided it defaults to *thetaM=270*.

    See :class:`FunctionalProfile` for a description of the the profile
    function.
    """

    RESERVED = ("profile", "tol", "name", "extent", "dead_below", "dead_above", "interface_below", "interface_above")
    magnetic = True

    def __init__(self, profile=None, tol=1e-3, name=None, **kw):
        if not name:
            name = profile.__name__
        if profile is None:
            raise TypeError("Need profile")
        # strip magnetism keywords from list of keywords
        magkw = dict((a, kw.pop(a)) for a in set(self.RESERVED) & set(kw.keys()))
        BaseMagnetism.__init__(self, name=name, **magkw)
        self.profile = profile
        self.tol = tol

        _set_parameters(self, name, profile, kw, self.RESERVED)
        rhoM_start = _MagnetismLimit(self, isend=False, isrhoM=True)
        rhoM_end = _MagnetismLimit(self, isend=True, isrhoM=True)
        thetaM_start = _MagnetismLimit(self, isend=False, isrhoM=False)
        thetaM_end = _MagnetismLimit(self, isend=True, isrhoM=False)
        self.start = Magnetism(rhoM=rhoM_start, thetaM=thetaM_start)
        self.end = Magnetism(rhoM=rhoM_end, thetaM=thetaM_end)
        self.anchor = None

    def set_anchor(self, stack, index):
        self.anchor = (stack, index)

    # TODO: is there a sane way of computing magnetism thickness in advance?
    def _calc_thickness(self):
        if self.anchor is None:
            raise ValueError("Need layer.magnetism.set_anchor(stack, layer) to compute magnetic thickness.")
        stack, index = self.anchor
        stack, start = stack._lookup(index)
        total = 0
        for k in range(start, start + self.extent):
            total += stack[k].thickness.value
        total -= self.dead_below.value + self.dead_above.value
        return total

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update(_get_parameters(self))
        return parameters

    def to_dict(self):
        ret = BaseMagnetism.to_dict(self)
        ret.update(
            to_dict(
                {
                    "profile": self.profile,
                    "parameters": _get_parameters(self),
                    "tol": self.tol,
                }
            )
        )

    def render(self, probe, slabs, thickness, anchor, sigma):
        Pw, Pz = slabs.microslabs(thickness)
        if len(Pw) == 0:
            return
        P = self.profile(Pz, **_get_values(self))

        rhoM, thetaM = P if isinstance(P, tuple) else (P, DEFAULT_THETA_M)
        try:
            # rhoM or thetaM may be constant, lists or arrays (but not tuples!)
            rhoM, thetaM = [broadcast_to(v, Pz.shape) for v in (rhoM, thetaM)]
        except ValueError:
            raise TypeError("profile function '%s' did not return array rhoM(z)" % self.profile.__name__)
        P = rhoM + thetaM * 0.001j  # combine rhoM/thetaM so they can be merged
        Pw, P = utils.merge_ends(Pw, P, tol=self.tol)
        rhoM, thetaM = P.real, P.imag * 1000  # split out rhoM,thetaM again
        slabs.add_magnetism(anchor=anchor, w=Pw, rhoM=rhoM, thetaM=thetaM, sigma=sigma)

    def __repr__(self):
        return "FunctionalMagnetism(%s)" % self.name


def _set_parameters(self, name, profile, kw, reserved):
    # Query profile function for the list of arguments
    vars = inspect.getfullargspec(profile)[0]
    # print "vars", vars
    if inspect.ismethod(profile):
        vars = vars[1:]  # Chop self
    vars = vars[1:]  # Chop z
    # print vars
    unused = [k for k in kw.keys() if k not in vars]
    if len(unused) > 0:
        raise TypeError("Profile got unexpected keyword argument '%s'" % unused[0])
    dups = [k for k in vars if k in reserved]
    if len(dups) > 0:
        raise TypeError("Profile has conflicting argument %r" % dups[0])
    for k in vars:
        kw.setdefault(k, 0)
    for k, v in kw.items():
        try:
            pv = [Parameter.default(vi, name=f"{name} {k}[{i}]") for i, vi in enumerate(v)]
        except TypeError:
            pv = Parameter.default(v, name=f"{name} {k}")
        setattr(self, k, pv)
    self._parameters = vars


def _get_parameters(self):
    return {k: getattr(self, k) for k in self._parameters}


def _get_values(self):
    vals = {}
    for k in self._parameters:
        v = getattr(self, k)
        if isinstance(v, list):
            vals[k] = asarray([vk.value for vk in v], "d")
        else:
            vals[k] = v.value
    return vals


class _LayerLimit(Calculation):
    def __init__(self, flayer, isend=True, isrho=True):
        self.description = f"{'rho' if isrho else 'irho'} Layer Limit, isend={isend}"
        self.flayer = flayer
        self.isend = isend
        self.isrho = isrho
        self.name = str(flayer) + self._tag

    @property
    def _tag(self):
        return (".rho_" if self.isrho else ".irho_") + ("end" if self.isend else "start")

    def parameters(self):
        return []

    def _function(self):
        z = asarray([0.0, self.flayer.thickness.value])
        P = self.flayer.profile(asarray(z), **_get_values(self.flayer))
        index = 1 if self.isend else 0
        return real(P[index]) if self.isrho else imag(P[index])

    def __repr__(self):
        return repr(self.flayer) + self._tag


class _MagnetismLimit(Calculation):
    def __init__(self, flayer, isend=True, isrhoM=True):
        self.description = f"{'rhoM' if isrhoM else 'thetaM'} Magnetism Limit, isend={isend}"
        self.flayer = flayer
        self.isend = isend
        self.isrhoM = isrhoM
        self.name = str(flayer) + self._tag

    @property
    def _tag(self):
        return (".rhoM_" if self.isrhoM else ".thetaM_") + ("end" if self.isend else "start")

    def parameters(self):
        return []

    def _function(self):
        zmax = self.flayer._calc_thickness()
        z = asarray([0.0, zmax])
        P = self.flayer.profile(z, **_get_values(self.flayer))
        rhoM, thetaM = P if isinstance(P, tuple) else (P, DEFAULT_THETA_M)
        rhoM, thetaM = [broadcast_to(v, z.shape) for v in (rhoM, thetaM)]
        index = -1 if self.isend else 0
        return rhoM[index] if self.isrhoM else thetaM[index]

    def __repr__(self):
        return repr(self.flayer) + self._tag



#!/usr/bin/env python
"""
Run bumps with refl1d plugin.

The *align* subcommand can be used on a completed DREAM fit to redraw the
profile contours aligned to a different layer boundary.
See :func:`refl1d.uncertainty.run_errors` for details.
"""

import sys

from . import __version__


def setup_bumps():
    """
    Install the refl1d plugin into bumps, but don't run main.
    """
    # vfs_init must be called pretty much first since it replaces filesystem
    # calls with filesystem hooks.  Any module that imports the calls directly
    # e.g., using *from os.path import exists*, needs to have the hook in
    # place before the module is even imported.
    try:
        from bumps.vfs import vfs_init

        vfs_init()
    except ImportError:
        # CRUFT: older bumps doesn't provide vfs
        pass
    import bumps.cli

    bumps.cli.set_mplconfig(appdatadir="Refl1D-" + __version__)
    from .bumps_interface import fitplugin

    bumps.cli.install_plugin(fitplugin)


def cli():
    """
    Install the Refl1D plugin into bumps and run the command line interface.
    """
    setup_bumps()

    # TODO: Add subcommand support to bumps.
    if len(sys.argv) > 2 and sys.argv[1] == "align":
        from .uncertainty import run_errors

        del sys.argv[1]
        run_errors()
    else:
        import bumps.cli

        bumps.cli.main()


def gui():
    """
    Install the Refl1D plugin into bumps and run the graphical user interface.
    """
    setup_bumps()
    import bumps.gui.gui_app

    bumps.gui.gui_app.main()


if __name__ == "__main__":
    cli()



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.contract_profile")

ALIGN_MAGNETIC_SIG = "i4(f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8[:,:])"
# @numba.njit(ALIGN_MAGNETIC_SIG, parallel=False, cache=True)
align_magnetic = numba.njit(cache=True)(MODULE.align_magnetic)
MODULE.align_magnetic = align_magnetic


CONTRACT_MAG_SIG = "i4(f8[:], f8[:], f8[:], f8[:], f8[:], f8[:], f8)"
# @numba.njit(CONTRACT_MAG_SIG, parallel=False, cache=True)
contract_mag = numba.njit(cache=True)(MODULE.contract_mag)
MODULE.contract_mag = contract_mag


CONTRACT_BY_AREA_SIG = "i4(f8[:], f8[:], f8[:], f8[:], f8)"
# @numba.njit(CONTRACT_BY_AREA_SIG, parallel=False, cache=True)
contract_by_area = numba.njit(cache=True)(MODULE.contract_by_area)
MODULE.contract_by_area = contract_by_area



"""
Resolution calculations
"""

from typing import TYPE_CHECKING

from numpy import (
    arange,
    asarray,
    ceil,
    cos,
    degrees,
    float64,
    hstack,
    isscalar,
    log,
    ones_like,
    pi,
    radians,
    sin,
    sqrt,
    tan,
)
from numpy import arcsin as asin

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def QL2T(Q=None, L=None):
    r"""
    Compute angle from $Q$ and wavelength.

    .. math::

        \theta = \sin^{-1}( |Q| \lambda / 4 \pi )

    Returns $\theta$\ |deg|.
    """
    Q, L = asarray(Q, "d"), asarray(L, "d")
    return degrees(asin(abs(Q) * L / (4 * pi)))


def QT2L(Q=None, T=None):
    r"""
    Compute wavelength from $Q$ and angle.

    .. math::

        \lambda = 4 \pi \sin( \theta )/Q

    Returns $\lambda$\ |Ang|.
    """
    Q, T = asarray(Q, "d"), radians(asarray(T, "d"))
    return 4 * pi * sin(T) / Q


def TL2Q(T: "ArrayLike", L: "ArrayLike"):
    r"""
    Compute $Q$ from angle and wavelength.

    .. math::

        Q = 4 \pi \sin(\theta) / \lambda

    Returns $Q$ |1/Ang|
    """
    T, L = radians(asarray(T, "d")), asarray(L, "d")
    return 4 * pi * sin(T) / L


_FWHM_scale: float64 = sqrt(log(256))


def FWHM2sigma(s: "ArrayLike"):
    return asarray(s, "d") / _FWHM_scale


def sigma2FWHM(s):
    return asarray(s, "d") * _FWHM_scale


def dTdL2dQ(T: "ArrayLike", dT: "ArrayLike", L: "ArrayLike", dL: "ArrayLike"):
    r"""
    Convert wavelength dispersion and angular divergence to $Q$ resolution.

    *T*, *dT*  (degrees) angle and FWHM angular divergence
    *L*, *dL*  (Angstroms) wavelength and FWHM wavelength dispersion

    Returns 1-\ $\sigma$ $\Delta Q$

    Given $Q = 4 \pi sin(\theta)/\lambda$, this follows directly from
    gaussian error propagation using

    ..math::

        \Delta Q^2
            &= \left(\frac{\partial Q}{\partial \lambda}\right)^2\Delta\lambda^2
            + \left(\frac{\partial Q}{\partial \theta}\right)^2\Delta\theta^2

            &= Q^2 \left(\frac{\Delta \lambda}{\lambda}\right)^2
            + Q^2 \left(\frac{\Delta \theta}{\tan \theta}\right)^2

            &= Q^2 \left(\frac{\Delta \lambda}{\lambda}\right)^2
            + \left(\frac{4\pi\cos\theta\,\Delta\theta}{\lambda}\right)^2

    with the final form chosen to avoid cancellation at $Q=0$.
    """

    # Compute dQ from wavelength dispersion (dL) and angular divergence (dT)
    T, dT = radians(asarray(T, float64)), radians(asarray(dT, float64))
    L, dL = asarray(L, float64), asarray(dL, float64)
    # print T, dT, L, dL
    dQ = (4 * pi / L) * sqrt((sin(T) * dL / L) ** 2 + (cos(T) * dT) ** 2)

    # sqrt((dL/L)**2+(radians(dT)/tan(radians(T)))**2)*probe.Q
    return FWHM2sigma(dQ)


def dQ_broadening(dQ, L, T, dT, width):
    r"""
    Broaden an existing dQ by the given divergence.

    *dQ* |1/Ang|, with 1-\ $\sigma$ $Q$ resolution
    *L* |Ang|
    *T*, *dT* |deg|, with FWHM angular divergence
    *width* |deg|, with FWHM increased angular divergence

    The calculation is derived by substituting
    $\Delta\theta' = \Delta\theta + \omega$ for sample broadening $\omega$
    into the resolution estimate
    $(\Delta Q/Q)^2 = (\Delta\lambda/\lambda)^2 + (\Delta\theta/\tan\theta)^2$.
    """
    T, dT = radians(asarray(T, "d")), FWHM2sigma(radians(asarray(dT, "d")))
    width = FWHM2sigma(radians(width))
    dQsq = dQ**2 + (4 * pi / L * cos(T)) ** 2 * (2 * width * dT + width**2)

    # If width < -dT, need to take abs(dQsq) before taking the sqrt
    # (focusing past zero)
    return sqrt(abs(dQsq))


def dQdT2dLoL(Q, dQ, T, dT):
    r"""
    Convert a calculated Q resolution and angular divergence to a
    wavelength dispersion.

    *Q*, *dQ* |1/Ang|  $Q$ and 1-\ $\sigma$ $Q$ resolution
    *T*, *dT* |deg| angle and FWHM angular divergence

    Returns FWHM $\Delta\lambda/\lambda$
    """
    T, dT = radians(asarray(T, "d")), radians(asarray(dT, "d"))
    Q, dQ = asarray(Q, "d"), asarray(dQ, "d")
    dQoQ = sigma2FWHM(dQ) / Q
    dToT = dT / tan(T)
    if (dQoQ < dToT).any():
        raise ValueError("Cannot infer wavelength resolution: dQ is too small or dT is too large for some data points")
    return sqrt(dQoQ**2 - dToT**2)


def dQdL2dT(Q, dQ, L, dL):
    r"""
    Convert a calculated Q resolution and wavelength dispersion to
    angular divergence.

    *Q*, *dQ* |1/Ang|  $Q$ and 1-\ $\sigma$ $Q$ resolution
    *L*, *dL* |deg| angle and FWHM angular divergence

    Returns FWHM \Delta\theta$
    """
    L, dL = asarray(L, "d"), asarray(dL, "d")
    Q, dQ = asarray(Q, "d"), asarray(dQ, "d")
    T = radians(QL2T(Q, L))
    dQoQ = sigma2FWHM(dQ) / Q
    dLoL = dL / L
    if (dQoQ < dLoL).any():
        raise ValueError("Cannot infer angular resolution: dQ is too small or dL is too large for some data points")
    dT = degrees(sqrt(dQoQ**2 - dLoL**2) * tan(T))
    return dT


Plancks_constant = 6.62618e-27  # Planck constant (erg*sec)
neutron_mass = 1.67495e-24  # neutron mass (g)


def TOF2L(d_moderator, TOF):
    r"""
    Convert neutron time-of-flight to wavelength.

    .. math::

        \lambda = (t/d) (h/n_m)

    where:

        | $\lambda$ is wavelength in |Ang|
        | $t$ is time-of-flight in $u$\s
        | $h$ is Planck's constant in erg seconds
        | $n_m$ is the neutron mass in g
    """
    return TOF * (Plancks_constant / neutron_mass / d_moderator)


def bins(low, high, dLoL):
    r"""
    Return bin centers from low to high preserving a fixed resolution.

    *low*, *high* are the minimum and maximum wavelength.
    *dLoL* is the desired resolution FWHM $\Delta\lambda/\lambda$ for the bins.
    """

    step = 1 + dLoL
    n = ceil(log(high / low) / log(step))
    edges = low * step ** arange(n + 1)
    L = (edges[:-1] + edges[1:]) / 2
    return L


def binwidths(L):
    r"""
    Determine the wavelength dispersion from bin centers *L*.

    The wavelength dispersion $\Delta\lambda$ is just the difference
    between consecutive bin edges, so:

    .. math::

        \Delta L_i  = E_{i+1}-E_{i}
                    = (1+\omega) E_i - E_i
                    = \omega E_i
                    = \frac{2 \omega}{2+\omega} L_i

    where $E$ and $\omega$ are as defined in :func:`binedges`.
    """
    L = asarray(L, "d")
    if L[1] > L[0]:
        dLoL = L[1] / L[0] - 1
    else:
        dLoL = L[0] / L[1] - 1
    dL = 2 * dLoL / (2 + dLoL) * L
    return dL


def binedges(L):
    r"""
    Construct bin edges *E* from bin centers *L*.

    Assuming fixed $\omega = \Delta\lambda/\lambda$ in the bins, the
    edges will be spaced logarithmically at:

    .. math::

        E_0     &= \min \lambda \\
        E_{i+1} &= E_i + \omega E_i = E_i (1+\omega)

    with centers $L$ half way between the edges:

    .. math::

        L_i = (E_i+E_{i+1})/2
            = (E_i + E_i (1+\omega))/2
            = E_i (2 + \omega)/2

    Solving for $E_i$, we can recover the edges from the centers:

    .. math::

        E_i = L_i \frac{2}{2+\omega}

    The final edge, $E_{n+1}$, does not have a corresponding center
    $L_{n+1}$ so we must determine it from the previous edge $E_n$:

    .. math::

        E_{n+1} = L_n \frac{2}{2+\omega}(1+\omega)

    The fixed $\omega$ can be retrieved from the ratio of any pair
    of bin centers using:

    .. math::

        \frac{L_{i+1}}{L_i} = \frac{ (E_{i+2}+E_{i+1})/2 }{ (E_{i+1}+E_i)/2 }
                          = \frac{ (E_{i+1}(1+\omega)+E_{i+1} }
                                  { (E_i(1+\omega)+E_i }
                          = \frac{E_{i+1}}{E_i}
                          = \frac{E_i(1+\omega)}{E_i} = 1 + \omega
    """
    L = asarray(L, "d")
    if L[1] > L[0]:
        dLoL = L[1] / L[0] - 1
        last = 1 + dLoL
    else:
        dLoL = L[0] / L[1] - 1
        last = 1.0 / (1 + dLoL)
    E = L * 2 / (2 + dLoL)
    return hstack((E, E[-1] * last))


def divergence(T=None, slits=None, distance=None, sample_width=1e10, sample_broadening=0):
    r"""
    Calculate divergence due to slit and sample geometry.

    :Parameters:
        *T*         : float OR [float] | degrees
            incident angles
        *slits*     : float OR (float, float) | mm
            s1, s2 slit openings for slit 1 and slit 2
        *distance*  : (float, float) | mm
            d1, d2 distance from sample to slit 1 and slit 2
        *sample_width*      : float | mm
            w, width of the sample
        *sample_broadening* : float | degrees FWHM
            additional divergence caused by sample

    :Returns:
        *dT*  : float OR [float] | degrees FWHM
            calculated angular divergence

    **Algorithm:**

    The divergence is based on the slit openings and the distance between
    the slits.  For very small samples, where the slit opening is larger
    than the width of the sample across the beam, the sample itself acts
    like the second slit.

    First find $p$, the projection of the beam on the sample:

    .. math::

        p &= w \sin\left(\frac{\pi}{180}\theta\right)

    Depending on whether $p$ is larger than $s_2$, determine the slit
    divergence $\Delta\theta_d$ in radians:

    .. math::

        \Delta\theta_d &= \left\{
          \begin{array}{ll}
            \frac{1}{2}\frac{s_1+s_2}{d_1-d_2} & \mbox{if } p \geq s_2 \\
            \frac{1}{2}\frac{s_1+p}{d_1}       & \mbox{if } p < s_2
          \end{array}
        \right.

    In addition to the slit divergence, we need to add in any sample
    broadening $\Delta\theta_s$ returning the total divergence in degrees:

    .. math::

        \Delta\theta &= \frac{180}{\pi} \Delta\theta_d + \Delta\theta_s

    Reversing this equation, the sample broadening contribution can
    be measured from the full width at half maximum of the rocking
    curve, $B$, measured in degrees at a particular angle and slit
    opening:

    .. math::

        \Delta\theta_s = B - \frac{180}{\pi}\Delta\theta_d
    """
    # TODO: update from reductus to handle four slits
    # TODO: check that the formula is correct for T=0 => dT = s1 / d1
    # TODO: add sample_offset and compute full footprint
    d1, d2 = distance
    try:
        s1, s2 = slits
    except TypeError:
        s1 = s2 = slits

    # Compute FWHM angular divergence dT from the slits in degrees
    dT = degrees(0.5 * (s1 + s2) / (d1 - d2))

    # For small samples, use the sample projection instead.
    sample_s = sample_width * sin(radians(T))
    if isscalar(sample_s):
        if sample_s < s2:
            dT = degrees(0.5 * (s1 + sample_s) / d1)
    else:
        idx = sample_s < s2
        # print s1, s2, d1, d2, T, dT, sample_s
        s1 = ones_like(sample_s) * s1
        dT = ones_like(sample_s) * dT
        dT[idx] = degrees(0.5 * (s1[idx] + sample_s[idx]) / d1)

    return dT + sample_broadening


def slit_widths(T=None, slits_at_Tlo=None, Tlo=90, Thi=90, slits_below=None, slits_above=None):
    """
    Compute the slit widths for the standard scanning reflectometer
    fixed-opening-fixed geometry.

    :Parameters:
        *T* : [float] | degrees
            Specular measurement angles.
        *Tlo*, *Thi* : float | degrees
            Start and end of the opening region.  The default if *Tlo* is
            not specified is to use fixed slits at *slits_below* for all
            angles.
        *slits_below*, *slits_above* : float OR [float, float] | mm
            Slits outside opening region.  The default is to use the
            values of the slits at the ends of the opening region.
        *slits_at_Tlo* : float OR [float, float] | mm
            Slits at the start of the opening region.

    :Returns:
        *s1*, *s2* : [float] | mm
            Slit widths for each theta.

    Slits are assumed to be fixed below angle *Tlo* and above angle *Thi*,
    and opening at a constant dT/T between them.

    Slit openings are defined by a tuple (s1, s2) or constant s=s1=s2.
    With no *Tlo*, the slits are fixed with widths defined by *slits_below*,
    which defaults to *slits_at_Tlo*.  With no *Thi*, slits are continuously
    opening above *Tlo*.

    .. Note::
         This function works equally well if angles are measured in
         radians and/or slits are measured in inches.

    """

    # Slits at T<Tlo
    if slits_below is None:
        slits_below = slits_at_Tlo
    try:
        b1, b2 = slits_below
    except TypeError:
        b1 = b2 = slits_below
    s1 = ones_like(T) * b1
    s2 = ones_like(T) * b2

    # Slits at Tlo<=T<=Thi
    try:
        m1, m2 = slits_at_Tlo
    except TypeError:
        m1 = m2 = slits_at_Tlo
    idx = abs(T) >= Tlo
    s1[idx] = m1 * T[idx] / Tlo
    s2[idx] = m2 * T[idx] / Tlo

    # Slits at T > Thi
    if slits_above is None:
        slits_above = m1 * Thi / Tlo, m2 * Thi / Tlo
    try:
        t1, t2 = slits_above
    except TypeError:
        t1 = t2 = slits_above
    idx = abs(T) > Thi
    s1[idx] = t1
    s2[idx] = t2

    return s1, s2


'''
def resolution(Q=None, s=None, d=None, L=None, dLoL=None, Tlo=None, Thi=None,
               s_below=None, s_above=None,
               broadening=0, sample_width=1e10, sample_distance=0):
    """
    Compute the resolution for Q on scanning reflectometers.

    broadening is the sample warp contribution to angular divergence, as
    measured by a rocking curve.  The value should be w - (s1+s2)/(2*d)
    where w is the full-width at half maximum of the rocking curve.

    For itty-bitty samples, provide a sample width w and sample distance ds
    from slit 2 to the sample.  If s_sample = sin(T)*w is smaller than s2
    for some T, then that will be used for the calculation of dT instead.

    """
    T = QL2T(Q=Q, L=L)
    slits = slit_widths(T=T, s=s, Tlo=Tlo, Thi=Thi)
    dT = divergence(T=T, slits=slits, sample_width=sample_width,
                    sample_distance=sample_distance) + broadening
    Q, dQ = Qresolution(L, dLoL*L, T, dT)
    return FWHM2sigma(dQ)


def demo():
    from numpy import linspace, exp, real, conj, sin, radians
    import matplotlib.pyplot as plt

    # Values from volfrac example in garefl
    T = linspace(0, 9, 140)
    Q = 4*pi*sin(radians(T))/5.0042
    dQ = resolution(Q, s=0.21, Tlo=0.35, d=1890., L=5.0042, dLoL=0.009)
    #plt.plot(Q, dQ)

    # Fresnel reflectivity for silicon
    rho, sigma=2.07, 5
    kz=Q/2
    f = sqrt(kz**2 - 4*pi*rho*1e-6 + 0j)
    r = (kz-f)/(kz+f)*exp(-2*sigma**2*kz*f)
    r[abs(kz)<1e-10] = -1
    R = real(r*conj(r))
    plt.errorbar(Q, R, xerr=dQ, fmt=',r', capsize=0)
    plt.grid(True)
    plt.semilogy(Q, R, ',b')

    plt.show()


def demo2():
    import numpy as np
    import matplotlib.pyplot as plt

    Q, R, dR = np.loadtxt('ga128.refl.mce').T
    dQ = resolution(Q, s=0.154, Tlo=0.36, d=1500., L=4.75, dLoL=0.02)
    plt.errorbar(Q, R, xerr=dQ, yerr=dR, fmt=',r', capsize=0)
    plt.grid(True)
    plt.semilogy(Q, R, ',b')
    plt.show()


if __name__ == "__main__":
    demo2()
'''



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.rebin")

# Define a bin iterator to adapt to either forward or reversed inputs.
spec = [
    ("forward", numba.boolean),
    ("n", numba.int64),
    ("edges", numba.float64[:]),
    ("bin", numba.int64),
    ("lo", numba.float64),
    ("hi", numba.float64),
    ("atend", numba.boolean),
]

BinIter = numba.experimental.jitclass(spec)(MODULE.BinIter)
MODULE.BinIter = BinIter

rebin_counts_portion = numba.njit(parallel=False, cache=True)(MODULE.rebin_counts_portion)
MODULE.rebin_counts_portion = rebin_counts_portion

rebin_counts = numba.njit(parallel=False, cache=True)(MODULE.rebin_counts)
MODULE.rebin_counts = rebin_counts

rebin_intensity = numba.njit(parallel=False, cache=True)(MODULE.rebin_intensity)

rebin_counts_2D = numba.njit(cache=True)(MODULE.rebin_counts_2D)



import numba
from .clone_module import clone_module

MODULE = clone_module("refl1d.lib.python.magnetic")

MODULE.prange = numba.prange

calculate_U1_U3_single = numba.njit(cache=True)(MODULE.calculate_U1_U3_single)
MODULE.calculate_U1_U3_single = calculate_U1_U3_single

calculate_u1_u3 = numba.njit(cache=True)(MODULE.calculate_u1_u3)
MODULE.calculate_u1_u3 = calculate_u1_u3


CR4XA_SIG = "void(i8, f8[:], f8[:], f8, f8[:], f8[:], f8[:], c16[:], c16[:], f8, i4, c16[:], c16[:], c16[:], c16[:])"
CR4XA_LOCALS = {
    "E0": numba.float64,
    "L": numba.int32,
    "LP": numba.int32,
    "STEP": numba.int8,
    "Z": numba.float64,
}
CR4XA_LOCALS.update(
    (s, numba.complex128)
    for s in [
        "S1L",
        "S1LP",
        "S3",
        "S3LP",
        "FS1S1",
        "FS1S3",
        "FS3S1",
        "FS3S3",
        "DELTA",
        "BL",
        "GL",
        "BLP",
        "GLP",
        "SSWAP",
        "DETW",
        "DBB",
        "DBG",
        "DGB",
        "DGG",
        "ES1L",
        "ENS1L",
        "ES1LP",
        "ENS1LP",
        "ES3L",
        "ENS3L",
        "ES3LP",
        "ENS3LP",
    ]
)
CR4XA_LOCALS.update(("A{i}{j}".format(i=i, j=j), numba.complex128) for i in range(1, 5) for j in range(1, 5))
CR4XA_LOCALS.update(("B{i}{j}".format(i=i, j=j), numba.complex128) for i in range(1, 5) for j in range(1, 5))
CR4XA_LOCALS.update(("C{i}".format(i=i), numba.complex128) for i in range(1, 5))

Cr4xa = numba.njit(CR4XA_SIG, parallel=False, cache=True, locals=CR4XA_LOCALS)(MODULE.Cr4xa)
MODULE.Cr4xa = Cr4xa


MAGAMP_SIG = "void(f8[:], f8[:], f8[:], f8[:], f8[:], c16[:], c16[:], f8[:], i4[:], c16[:], c16[:], c16[:], c16[:])"

magnetic_amplitude = numba.njit(MAGAMP_SIG, parallel=False, cache=True)(MODULE.magnetic_amplitude)



"""
Reflectometry profile interactor.
"""

import math

import numpy as np
from matplotlib import transforms
from numpy import inf

from . import registry
from .binder import BindArtist, pixel_to_data
from .config import layer_hysteresis, rho_color, rhoI_color, rhoM_color, thetaM_color
from .interactor import BaseInteractor, safecall
from .interface import InterfaceInteractor
from .thickness import ThicknessInteractor

blend_xy = transforms.blended_transform_factory


# ================== Main profile interactor =====================
class ProfileInteractor(object):
    """
    Reflectometry profile editor
    """

    _debug = False

    def __init__(self, axes, theta_axes, status_update):
        self.status_update = status_update
        self.canvas = axes.figure.canvas
        self.axes = axes
        self.theta_axes = theta_axes
        self.xcoords = blend_xy(axes.transData, axes.transAxes)

        # Add some plots
        [self.hrho] = axes.plot([], [], "-", color=rho_color, label=r"$\rho$")
        [self.hrhoI] = axes.plot([], [], "-", color=rhoI_color, label=r"$\rho_i$")
        [self.hrhoM] = self.axes.plot([], [], "-", color=rhoM_color, label=r"$\rho_M$")
        [self.hthetaM] = self.theta_axes.plot([], [], "-", color=thetaM_color, label=r"$\theta_M$")

        # Ick! trying to do motion event before profile has been set
        self.experiment = None

        # TODO: the connect mechanism needs to be owned by the canvas rather
        # than the axes --- cannot have multiple profiles on the same canvas
        # until connect is in the right place.
        self.connect = BindArtist(axes.figure)
        self.connect.clearall()
        self.connect("motion", axes, self.onMotion)
        self.connect("click", axes.figure, self.onClick)

        # Set interactors
        self.thickness_interactor = ThicknessInteractor(self)
        self.interface_interactor = InterfaceInteractor(self)
        self.layer_interactor = BaseInteractor(self)

    def set_experiment(self, experiment, force_recalc, signal_update):
        self.experiment = experiment
        self.force_recalc = force_recalc
        self.signal_update = signal_update

        # Theta needs a separate axis, we put these two axes into a figure
        self.axes.set_xlabel(r"$\rm{z}/\AA$")
        self.ismagnetic = experiment.sample.ismagnetic
        self.isstep = experiment.step_interfaces
        if self.ismagnetic:
            self.axes.set_ylabel(r"$\rm{SLD}\ \rho,\  \rho_i,\  \rho_M\ /\ 10^{-6} \AA^{-2}$")
            self.theta_axes.set_ylabel(r"$\rm{Magnetic\ Angle}\ \theta_M\ /\ ^\circ$")
            self.theta_axes.set_visible(True)
            self.hrhoM.set_visible(True)
            self.hthetaM.set_visible(True)
            self.hlegend = self.axes.legend(
                (self.hrho, self.hrhoI, self.hrhoM, self.hthetaM),
                # ('SLD','Absorption','Mag. SLD','Mag. angle'),
                (r"$\rho$", r"$\rho_i$", r"$\rho_M$", r"$\theta_M$"),
                loc=(0.85, 0.5),
                # loc='upper right'
            )
        else:
            self.axes.set_ylabel(r"$\rm{SLD}\ \rho,\  \rho_i\ /\ 10^{-6} \AA^{-2}$")
            self.theta_axes.set_ylabel("")
            self.theta_axes.set_visible(False)
            self.hrhoM.set_visible(False)
            self.hthetaM.set_visible(False)
            self.hlegend = self.axes.legend(
                (self.hrho, self.hrhoI),
                (r"$\rho$", r"$\rho_i$"),
                loc=(0.85, 0.5),
                # loc='upper right'
            )
        self.hlegend.get_frame().set(alpha=0.2, facecolor="yellow")

        self.x_offset = 0
        self.layer_num = None
        self._find_layer_boundaries()
        self.thickness_interactor.reset_markers()
        self.set_layer(0)
        self.redraw()

    def update_cursor(self, x, y):
        def nice(value, range):
            place = int(math.log10(abs(range[1] - range[0])) - 3)
            # print value,range,place
            if place < 0:
                return "%.*f" % (-place, value)
            else:
                return "%d" % int(value)

        self.status_update("x:%s  y:%s" % (nice(x, self.axes.get_xlim()), nice(y, self.axes.get_ylim())))

    def onMotion(self, event):
        """Respond to motion events by changing the active layer."""
        if self.experiment is None:
            return False

        # Force data coordinates for the mouse position
        transform = self.axes.transData
        x, y = pixel_to_data(transform, event.x, event.y)
        self.update_cursor(x, y)

        # Check if mouse is still in the same layer
        if self.layer_num is not None:
            left, right = self.boundary[self.layer_num : self.layer_num + 2]
            # Quick rejection if it is in the layer
            if left < x < right:
                # print "quick reject", event.xdata, self.layer_start, self.layer_end
                return False
            # If it is within 5 pixels of the layer, keep to the same layer
            xlo, _ = pixel_to_data(transform, event.x + layer_hysteresis, event.y)
            xhi, _ = pixel_to_data(transform, event.x - layer_hysteresis, event.y)
            if left < xlo and xhi < right:
                # print "hysteresis", event.x, xlo, xhi, self.layer_start, self.layer_end
                return False

        # If layer changed, then replace the markers
        self.set_layer(x)
        return False

    def onClick(self, ev):
        # Couldn't make popups work --- moved context menu to panel
        return False

    # === Sample information ===
    def find(self, z):
        return self.experiment.sample.find(z - self.x_offset)

    def _find_layer_boundaries(self):
        offset = self.x_offset
        boundary = [-inf, offset]
        if hasattr(self.experiment, "sample"):
            for L in self.experiment.sample[1:-1]:
                dx = L.thickness.value
                offset += dx
                boundary.append(offset)
        boundary.append(inf)
        self.boundary = np.asarray(boundary)

    def sample_labels(self):
        return [L.name for L in self.experiment.sample]

    def sample_layer(self, n):
        return self.experiment.sample[n]

    @safecall
    def set_layer(self, z):
        """Make layer containing z the active layer."""
        # Note: sample_layer needs layer_num to be an int, rather than a
        # numpy int64, since it tests for int vs. string vs. material.
        self.layer_num = int(np.searchsorted(self.boundary, z) - 1)
        layer = self.sample_layer(self.layer_num)

        # Reset the interface markers
        self.interface_interactor.set_layer()

        # Reset the layer markers
        self.layer_interactor.clear_markers()
        self.layer_interactor = registry.interactor(self, layer)

        # Update the marker positions
        self.update_markers()

        # Trigger redraw
        self.draw_idle()

    def update_markers(self):
        self._find_layer_boundaries()
        self.thickness_interactor.update_markers()
        self.interface_interactor.update_markers()
        self.layer_interactor.update_markers()

    def update_profile(self):
        if self.ismagnetic:
            z, rho, rhoI, rhoM, thetaM = self.experiment.magnetic_smooth_profile()
            self.hrhoM.set_data(z, rhoM)
            self.hthetaM.set_data(z, thetaM)
        else:
            z, rho, rhoI = self.experiment.smooth_profile()
        # print "z",z
        # print "offset",self.x_offset
        self.hrho.set_data(z + self.x_offset, rho)
        self.hrhoI.set_data(z + self.x_offset, rhoI)

    def reset_limits(self):
        # Compute automatic y limits
        # Note: theta limits are on ax2

        # TODO: the markers themselves may fall outside the graph.
        # May need to walk sample to find the position of the markers
        # when setting the limits.  Alternatively, we could just ask
        # where the markers are.
        z = self.hrho.get_xdata()
        rho = self.hrho.get_ydata()
        rhoI = self.hrhoI.get_ydata()
        self.axes.set_xlim(z[0], z[-1])

        if self.ismagnetic:
            rhoM = self.hrhoM.get_ydata()
            thetaM = self.hthetaM.get_ydata()
            lo = min(rho.min(), rhoI.min(), rhoM.min())
            hi = max(rho.max(), rhoI.max(), rhoM.max())
            fluff = 0.05 * (hi - lo)
            self.axes.set_ylim(lo - fluff, hi + fluff)

            lo, hi = thetaM.min(), thetaM.max()
            fluff = 0.05 * (hi - lo)
            if fluff <= 1e-3:
                lo, hi, fluff = 0, 360, 0
            self.theta_axes.set_ylim(lo - fluff, hi + fluff)

        else:
            lo = min(rho.min(), rhoI.min())
            hi = max(rho.max(), rhoI.max())
            fluff = 0.05 * (hi - lo)
            self.axes.set_ylim(lo - fluff, hi + fluff)

    def update(self):
        """
        Respond to changes in the model by recalculating the profiles and
        resetting the widgets.
        """
        # We are done the manipulation; let the model send its update signal
        # to whomever is listening.
        self.force_recalc()
        self.redraw()
        self.signal_update()

    def redraw(self, reset_limits=False):
        self.update_markers()
        self.update_profile()
        if reset_limits:
            self.reset_limits()
        self.draw_now()

    def draw_now(self):
        # print "draw immediately"
        self.canvas.draw()

    def draw_idle(self):
        """Set the limits and tell the canvas to render itself."""
        # print "draw when idle"
        self.canvas.draw_idle()
